"""Objects for ipranges."""

__all__ = ["IPRanges", "IPRange"]

from . import check, Object, ObjectField, ObjectFieldRelated, ObjectSet, ObjectType, to
from .subnets import Subnet
from ..enum import IPRangeType
from typing import Union

TYPE = type


class IPRangesType(ObjectType):
    """Metaclass for `IPRanges`."""

    async def read(cls):
        data = await cls._handler.read()
        return cls(map(cls._object, data))

    async def create(
        cls,
        start_ip: str,
        end_ip: str,
        *,
        type: IPRangeType = IPRangeType.RESERVED,
        comment: str = None,
        subnet: Union[Subnet, int] = None
    ):
        """
        Create a `IPRange` in MAAS.

        :param start_ip: First IP address in the range (required).
        :type start_ip: `str`
        :parma end_ip: Last IP address in the range (required).
        :type end_ip: `str`
        :param type: Type of IP address range (optional).
        :type type: `IPRangeType`
        :param comment: Reason for the IP address range (optional).
        :type comment: `str`
        :param subnet: Subnet the IP address range should be created on
            (optional). By default MAAS will calculate the correct subnet
            based on the `start_ip` and `end_ip`.
        :type subnet: `Subnet` or `int`
        :returns: The created IPRange
        :rtype: `IPRange`
        """
        if not isinstance(type, IPRangeType):
            raise TypeError("type must be an IPRangeType, not %s" % TYPE(type).__name__)

        params = {"start_ip": start_ip, "end_ip": end_ip, "type": type.value}
        if comment is not None:
            params["comment"] = comment
        if subnet is not None:
            if isinstance(subnet, Subnet):
                params["subnet"] = subnet.id
            elif isinstance(subnet, int):
                params["subnet"] = subnet
            else:
                raise TypeError(
                    "subnet must be Subnet or int, not %s" % (TYPE(subnet).__class__)
                )
        return cls._object(await cls._handler.create(**params))


class IPRanges(ObjectSet, metaclass=IPRangesType):
    """The set of IPRanges stored in MAAS."""


class IPRangeTypeMeta(ObjectType):
    """Metaclass for `IPRange`."""

    async def read(cls, id: int):
        """Get a `IPRange` by its `id`."""
        data = await cls._handler.read(id=id)
        return cls(data)


class IPRange(Object, metaclass=IPRangeTypeMeta):
    """A IPRange."""

    id = ObjectField.Checked("id", check(int), readonly=True, pk=True)
    start_ip = ObjectField.Checked("start_ip", check(str))
    end_ip = ObjectField.Checked("end_ip", check(str))
    type = ObjectField.Checked("type", to(IPRangeType), readonly=True)
    comment = ObjectField.Checked("comment", check(str))
    subnet = ObjectFieldRelated("subnet", "Subnet", readonly=True, pk=0)

    async def delete(self):
        """Delete this IPRange."""
        await self._handler.delete(id=self.id)



"""Test for `maas.client.viscera.events`."""

from datetime import datetime
from itertools import chain, count
import random
from unittest.mock import sentinel

from testtools.matchers import Equals, IsInstance

from ..users import User
from .. import events
from ...testing import make_mac_address, make_name_without_spaces, make_range, TestCase
from ..testing import bind


event_ids = count(1)


def make_Event_dict():
    return {
        "id": next(event_ids),
        "type": make_name_without_spaces("event-type"),
        "node": random.randint(1, 99),
        "hostname": make_name_without_spaces("host"),
        "level": random.choice(list(events.Level)),
        "created": datetime.utcnow().strftime("%a, %d %b. %Y %H:%M:%S"),
        "description": make_name_without_spaces("description"),
        "username": make_name_without_spaces("username"),
    }


def make_origin():
    # Create a new origin with Events and Event. The former refers to the
    # latter via the origin, hence why it must be bound.
    origin = bind(events.Events, events.Event)
    # Prepare a sensible default response from Events.query.
    events_query = origin.session.handlers[events.Events.__name__].query
    events_query.return_value = make_queried_events()
    return origin


def make_queried_events():
    """Mimic the object returned from a query."""
    return {"events": [], "prev_uri": sentinel.prev_uri, "next_uri": sentinel.next_uri}


class TestEventsQuery(TestCase):
    """Tests for `Events.query`."""

    def test__query_without_arguments_results_in_empty_bones_query(self):
        obj = make_origin().Events
        obj.query()
        obj._handler.query.assert_called_once_with()

    def test__query_arguments_are_assembled_and_passed_to_bones_handler1(self):
        obj = make_origin().Events
        arguments = {
            "hostnames": (
                make_name_without_spaces("hostname"),
                make_name_without_spaces("hostname"),
            ),
            "domains": (
                make_name_without_spaces("domain"),
                make_name_without_spaces("domain"),
            ),
            "zones": (
                make_name_without_spaces("zone"),
                make_name_without_spaces("zone"),
            ),
            "macs": (make_mac_address(), make_mac_address()),
            "system_ids": (
                make_name_without_spaces("system-id"),
                make_name_without_spaces("system-id"),
            ),
            "agent_name": make_name_without_spaces("agent"),
            "level": random.choice(list(events.Level)),
            "limit": random.randrange(1, 1000),
            "owner": make_name_without_spaces("username"),
        }
        obj.query(**arguments)
        expected = {
            "hostname": list(arguments["hostnames"]),
            "domain": list(arguments["domains"]),
            "zone": list(arguments["zones"]),
            "mac_address": list(arguments["macs"]),
            "id": list(arguments["system_ids"]),
            "agent_name": [arguments["agent_name"]],
            "level": [arguments["level"].name],
            "limit": [str(arguments["limit"])],
            "owner": [arguments["owner"]],
        }
        obj._handler.query.assert_called_once_with(**expected)

    def test__query_arguments_are_assembled_and_passed_to_bones_handler2(self):
        obj = make_origin().Events
        user = User(
            {
                "username": make_name_without_spaces("username"),
                "email": make_name_without_spaces("user@"),
                "is_superuser": False,
            }
        )
        arguments = {
            "hostnames": (
                make_name_without_spaces("hostname"),
                make_name_without_spaces("hostname"),
            ),
            "domains": (
                make_name_without_spaces("domain"),
                make_name_without_spaces("domain"),
            ),
            "zones": (
                make_name_without_spaces("zone"),
                make_name_without_spaces("zone"),
            ),
            "macs": (make_mac_address(), make_mac_address()),
            "system_ids": (
                make_name_without_spaces("system-id"),
                make_name_without_spaces("system-id"),
            ),
            "agent_name": make_name_without_spaces("agent"),
            "level": random.choice(list(events.Level)),
            "limit": random.randrange(1, 1000),
            "owner": user,
        }
        obj.query(**arguments)
        expected = {
            "hostname": list(arguments["hostnames"]),
            "domain": list(arguments["domains"]),
            "zone": list(arguments["zones"]),
            "mac_address": list(arguments["macs"]),
            "id": list(arguments["system_ids"]),
            "agent_name": [arguments["agent_name"]],
            "level": [arguments["level"].name],
            "limit": [str(arguments["limit"])],
            "owner": [user.username],
        }
        obj._handler.query.assert_called_once_with(**expected)

    def test__query_arguments_raises_error_for_invalid_owner(self):
        obj = make_origin().Events
        arguments = {
            "hostnames": (
                make_name_without_spaces("hostname"),
                make_name_without_spaces("hostname"),
            ),
            "domains": (
                make_name_without_spaces("domain"),
                make_name_without_spaces("domain"),
            ),
            "zones": (
                make_name_without_spaces("zone"),
                make_name_without_spaces("zone"),
            ),
            "macs": (make_mac_address(), make_mac_address()),
            "system_ids": (
                make_name_without_spaces("system-id"),
                make_name_without_spaces("system-id"),
            ),
            "agent_name": make_name_without_spaces("agent"),
            "level": random.choice(list(events.Level)),
            "limit": random.randrange(1, 1000),
            "owner": random.randint(0, 10),
        }
        self.assertRaises(TypeError, obj.query, **arguments)

    def test__query_level_is_normalised(self):
        obj = make_origin().Events
        for level in events.Level:
            for value in (level, level.name, level.value):
                obj._handler.query.reset_mock()
                obj.query(level=value)
                obj._handler.query.assert_called_once_with(level=[level.name])

    def test__query_before_argument_is_passed_to_bones_handler(self):
        obj = make_origin().Events
        before = random.randint(1, 1000)
        obj.query(before=before)
        obj._handler.query.assert_called_once_with(before=[str(before)])

    def test__query_after_argument_is_passed_to_bones_handler(self):
        obj = make_origin().Events
        after = random.randint(1, 1000)
        obj.query(after=after)
        obj._handler.query.assert_called_once_with(after=[str(after)])

    def test__cannot_query_with_both_before_and_after(self):
        self.assertRaises(ValueError, events.Events.query, before=2, after=1)


class TestEvents(TestCase):
    """Tests for `Events`."""

    def test__prev_requests_page_of_older_events(self):
        obj = make_origin().Events
        evts = obj(
            {
                "events": [],
                "prev_uri": "endpoint?before=100&limit=20&foo=abc",
                "next_uri": "endpoint?after=119&limit=20&foo=123",
            }
        )
        self.assertThat(evts.prev(), IsInstance(events.Events))
        evts._handler.query.assert_called_once_with(
            before=["100"], limit=["20"], foo=["abc"]
        )

    def test__next_requests_page_of_newer_events(self):
        obj = make_origin().Events
        evts = obj(
            {
                "events": [],
                "prev_uri": "endpoint?before=100&limit=20&foo=abc",
                "next_uri": "endpoint?after=119&limit=20&foo=123",
            }
        )
        self.assertThat(evts.next(), IsInstance(events.Events))
        evts._handler.query.assert_called_once_with(
            after=["119"], limit=["20"], foo=["123"]
        )

    def test__forwards_returns_a_continuous_iterator(self):
        pages = [
            {
                "events": [make_Event_dict() for _ in make_range()],
                "prev_uri": "?going=backwards",
                "next_uri": "?going=forwards",
            },
            {
                "events": [make_Event_dict() for _ in make_range()],
                "prev_uri": "?going=backwards",
                "next_uri": "?going=forwards",
            },
            # An empty page is taken to mean "stop".
            {
                "events": [],
                "prev_uri": "?going=backwards",
                "next_uri": "?going=forwards",
            },
        ]
        obj = make_origin().Events
        obj._handler.query.side_effect = pages
        self.assertThat(
            [evt._data for evt in obj.query().forwards()],
            Equals(
                list(chain.from_iterable(reversed(page["events"]) for page in pages))
            ),
        )
        # The query parameters in next_uri get passed through to bones.
        obj._handler.query.assert_called_with(going=["forwards"])

    def test__backwards_returns_a_continuous_iterator(self):
        pages = [
            {
                "events": [make_Event_dict() for _ in make_range()],
                "prev_uri": "?going=backwards",
                "next_uri": "?going=forwards",
            },
            {
                "events": [make_Event_dict() for _ in make_range()],
                "prev_uri": "?going=backwards",
                "next_uri": "?going=forwards",
            },
            # An empty page is taken to mean "stop".
            {
                "events": [],
                "prev_uri": "?going=backwards",
                "next_uri": "?going=forwards",
            },
        ]
        obj = make_origin().Events
        obj._handler.query.side_effect = pages
        self.assertThat(
            [evt._data for evt in obj.query().backwards()],
            Equals(list(chain.from_iterable(page["events"] for page in pages))),
        )
        # The query parameters in prev_uri get passed through to bones.
        obj._handler.query.assert_called_with(going=["backwards"])



"""Objects for devices."""

__all__ = ["Device", "Devices"]

import typing

from .nodes import Node, Nodes, NodesType, NodeTypeMeta
from .zones import Zone


class DevicesType(NodesType):
    """Metaclass for `Devices`."""

    async def create(
        cls,
        mac_addresses: typing.Sequence[str],
        hostname: str = None,
        domain: typing.Union[int, str] = None,
        zone: typing.Union[str, Zone] = None,
    ):
        """Create a new device.

        :param mac_addresses: The MAC address(es) of the device (required).
        :type mac_addresses: sequence of `str`
        :param hostname: The hostname for the device (optional).
        :type hostname: `str`
        :param domain: The domain for the device (optional).
        :type domain: `int` or `str`
        :param zone: The zone for the device (optional).
        :type zone: `Zone` or `str`

        """
        params = {"mac_addresses": mac_addresses}
        if hostname is not None:
            params["hostname"] = hostname
        if domain is not None:
            params["domain"] = domain
        if zone is not None:
            if isinstance(zone, Zone):
                params["zone"] = zone.name
            elif isinstance(zone, str):
                params["zone"] = zone
            else:
                raise TypeError(
                    "zone must be a str or Zone, not %s" % type(zone).__name__
                )
        return cls._object(await cls._handler.create(**params))


class Devices(Nodes, metaclass=DevicesType):
    """The set of devices stored in MAAS."""


class DeviceType(NodeTypeMeta):
    """Metaclass for `Device`."""


class Device(Node, metaclass=DeviceType):
    """A device stored in MAAS."""



"""Test for `maas.client.viscera.ipranges`."""

import random

from testtools.matchers import Equals

from ..ipranges import IPRange, IPRanges

from ..testing import bind
from ...enum import IPRangeType
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with IPRanges and IPRange. The former
    refers to the latter via the origin, hence why it must be bound.
    """
    return bind(IPRanges, IPRange)


class TestIPRanges(TestCase):
    def test__ipranges_create(self):
        IPRanges = make_origin().IPRanges
        start_ip = make_string_without_spaces()
        end_ip = make_string_without_spaces()
        type = IPRangeType.DYNAMIC
        comment = make_string_without_spaces()
        IPRanges._handler.create.return_value = {
            "id": 1,
            "start_ip": start_ip,
            "end_ip": end_ip,
            "type": type.value,
            "comment": comment,
        }
        IPRanges.create(start_ip=start_ip, end_ip=end_ip, type=type, comment=comment)
        IPRanges._handler.create.assert_called_once_with(
            start_ip=start_ip, end_ip=end_ip, type=type.value, comment=comment
        )

    def test__ipranges_create_requires_IPRangeType(self):
        IPRanges = make_origin().IPRanges
        start_ip = make_string_without_spaces()
        end_ip = make_string_without_spaces()
        comment = make_string_without_spaces()
        error = self.assertRaises(
            TypeError,
            IPRanges.create,
            start_ip=start_ip,
            end_ip=end_ip,
            type=make_string_without_spaces(),
            comment=comment,
        )
        self.assertEquals("type must be an IPRangeType, not str", str(error))

    def test__ipranges_read(self):
        """IPRanges.read() returns a list of IPRanges."""
        IPRanges = make_origin().IPRanges
        ipranges = [
            {
                "id": random.randint(0, 100),
                "start_ip": make_string_without_spaces(),
                "end_ip": make_string_without_spaces(),
                "type": make_string_without_spaces(),
                "comment": make_string_without_spaces(),
            }
            for _ in range(3)
        ]
        IPRanges._handler.read.return_value = ipranges
        ipranges = IPRanges.read()
        self.assertThat(len(ipranges), Equals(3))


class TestIPRange(TestCase):
    def test__iprange_read(self):
        IPRange = make_origin().IPRange
        iprange = {
            "id": random.randint(0, 100),
            "start_ip": make_string_without_spaces(),
            "end_ip": make_string_without_spaces(),
            "type": make_string_without_spaces(),
            "comment": make_string_without_spaces(),
        }
        IPRange._handler.read.return_value = iprange
        self.assertThat(IPRange.read(id=iprange["id"]), Equals(IPRange(iprange)))
        IPRange._handler.read.assert_called_once_with(id=iprange["id"])

    def test__iprange_delete(self):
        IPRange = make_origin().IPRange
        iprange_id = random.randint(1, 100)
        iprange = IPRange(
            {
                "id": iprange_id,
                "start_ip": make_string_without_spaces(),
                "end_ip": make_string_without_spaces(),
                "type": make_string_without_spaces(),
                "comment": make_string_without_spaces(),
            }
        )
        iprange.delete()
        IPRange._handler.delete.assert_called_once_with(id=iprange_id)



"""Testing helpers for `maas.client.utils`."""

from ...testing import make_name_without_spaces
from ..creds import Credentials


def make_Credentials():
    return Credentials(
        make_name_without_spaces("consumer_key"),
        make_name_without_spaces("token_key"),
        make_name_without_spaces("secret_key"),
    )



""" Testing framework for maas.client.viscera """

__all__ = ["bind"]


from collections.abc import Mapping
from itertools import chain
from unittest.mock import Mock

from .. import OriginBase
from ...testing import AsyncCallableMock


def bind(*objects, session=None):
    """Bind `objects` into a new `Origin` derived from `session`.

    The origin is constructed using `maas.client.viscera.OriginBase` hence
    the only other objects that one object may refer to are those given in
    `objects`. It's exactly like a minimally populated `Origin`.

    If provided, `session` is used when constructing the `OriginBase`
    instance, otherwise a mock is substituted. This mock has an empty
    `handlers` mapping.

    :param objects: Any number of `Object` classes, or mappings of `Object`
        classes from the names with which they should be bound in the origin
        that's created and returned.
    :param session: A `bones.SessionAPI` instance.
    :return: An `OriginBase` instance.
    """

    def _flatten_to_items(thing):
        if isinstance(thing, Mapping):
            yield from thing.items()
        else:
            yield thing.__name__, thing

    objects = map(_flatten_to_items, objects)
    objects = chain.from_iterable(objects)
    objects = dict(objects)

    if session is None:
        session = Mock(name="session")
        session.handlers = {
            name: AsyncCallableMock(name="handler(%s)" % name) for name in objects
        }

    return OriginBase(session, objects=objects)



"""Commands for devices."""

__all__ = ["register"]

from . import CommandError, OriginPagedTableCommand, tables


class cmd_devices(OriginPagedTableCommand):
    """List devices."""

    def __init__(self, parser):
        super(cmd_devices, self).__init__(parser)
        parser.add_argument("hostname", nargs="*", help=("Hostname of the device."))
        parser.add_argument(
            "--owned", action="store_true", help=("Show only machines owned by you.")
        )

    def execute(self, origin, options, target):
        hostnames = None
        if options.hostname:
            hostnames = options.hostname
        devices = origin.Devices.read(hostnames=hostnames)
        if options.owned:
            me = origin.Users.whoami()
            devices = origin.Devices(
                [
                    device
                    for device in devices
                    if device.owner is not None and device.owner.username == me.username
                ]
            )
        table = tables.DevicesTable()
        return table.render(target, devices)


class cmd_device(OriginPagedTableCommand):
    """Details of a device."""

    def __init__(self, parser):
        super(cmd_device, self).__init__(parser)
        parser.add_argument("hostname", nargs=1, help=("Hostname of the device."))

    def execute(self, origin, options, target):
        devices = origin.Devices.read(hostnames=options.hostname)
        if len(devices) == 0:
            raise CommandError("Unable to find device %s." % options.hostname[0])
        device = devices[0]
        table = tables.DeviceDetail()
        return table.render(target, device)


def register(parser):
    """Register commands with the given parser."""
    cmd_devices.register(parser)
    cmd_device.register(parser)



"""Integration tests for `maas.client`."""

from collections import Mapping
from datetime import datetime
from http import HTTPStatus
import io
from itertools import repeat
import random
from time import sleep

from maas.client import (
    bones,
    viscera,
)
from maas.client.testing import (
    make_name_without_spaces,
    TestCase,
)
from maas.client.utils import (
    creds,
    profiles,
    retries,
)
from testtools.matchers import (
    AllMatch,
    Equals,
    Is,
    IsInstance,
    MatchesAll,
    MatchesAny,
    MatchesStructure,
)


kiB = 2 ** 10
MiB = 2 ** 20


def scenarios():
    with profiles.ProfileStore.open() as config:
        return tuple(
            (profile.name, dict(profile=profile))
            for profile in map(config.load, config)
        )


class IntegrationTestCase(TestCase):

    scenarios = scenarios()

    def setUp(self):
        super(IntegrationTestCase, self).setUp()
        self.session = bones.SessionAPI.fromProfile(self.profile)
        self.origin = viscera.Origin(self.session)


class TestAccount(IntegrationTestCase):

    def test__create_and_delete_credentials(self):
        credentials = self.origin.Account.create_credentials()
        self.assertThat(credentials, IsInstance(creds.Credentials))
        self.origin.Account.delete_credentials(credentials)


class TestBootResources(IntegrationTestCase):

    def test__list_boot_resources(self):
        boot_resources = self.origin.BootResources.read()
        self.assertThat(boot_resources, MatchesAll(
            IsInstance(self.origin.BootResources),
            AllMatch(IsInstance(self.origin.BootResource)),
        ))
        self.assertThat(
            boot_resources,
            AllMatch(MatchesStructure(
                id=IsInstance(int),
                type=IsInstance(str),
                name=IsInstance(str),
                architecture=IsInstance(str),
                subarches=Optional(IsInstance(str)),
                sets=Optional(IsInstance(Mapping)),
            )),
        )

    def test__create_and_delete_boot_resource(self):
        chunk = random.getrandbits(8 * 128).to_bytes(128, "big")
        content = b"".join(repeat(chunk, 5 * MiB // len(chunk)))
        boot_resource = self.origin.BootResources.create(
            make_name_without_spaces("ubuntu", "/"), "amd64/generic",
            io.BytesIO(content))
        self.assertThat(boot_resource, IsInstance(self.origin.BootResource))
        boot_resource.delete()
        error = self.assertRaises(
            bones.CallError, self.origin.BootResource.read, boot_resource.id)
        self.assertThat(error, MatchesStructure(
            status=Equals(HTTPStatus.NOT_FOUND)))


class TestBootSources(IntegrationTestCase):

    def test__create_and_delete_source_with_keyring_filename(self):
        source_url = make_name_without_spaces("http://maas.example.com/")
        keyring_filename = make_name_without_spaces("keyring-filename")
        boot_source = self.origin.BootSources.create(
            source_url, keyring_filename=keyring_filename)
        self.assertThat(boot_source, IsInstance(self.origin.BootSource))
        boot_source.delete()
        error = self.assertRaises(
            bones.CallError, self.origin.BootSource.read, boot_source.id)
        self.assertThat(error, MatchesStructure(
            status=Equals(HTTPStatus.NOT_FOUND)))

    def test__create_and_delete_source_with_keyring_data(self):
        source_url = make_name_without_spaces("http://maas.example.com/")
        keyring_data = make_name_without_spaces("keyring-data").encode()
        boot_source = self.origin.BootSources.create(
            source_url, keyring_data=io.BytesIO(keyring_data))
        self.assertThat(boot_source, IsInstance(self.origin.BootSource))
        boot_source.delete()
        error = self.assertRaises(
            bones.CallError, self.origin.BootSource.read, boot_source.id)
        self.assertThat(error, MatchesStructure(
            status=Equals(HTTPStatus.NOT_FOUND)))

    def test__list_boot_sources(self):
        boot_sources = self.origin.BootSources.read()
        self.assertThat(boot_sources, MatchesAll(
            IsInstance(self.origin.BootSources),
            AllMatch(IsInstance(self.origin.BootSource)),
        ))
        self.assertThat(
            boot_sources,
            AllMatch(MatchesStructure(
                id=IsInstance(int),
                url=IsInstance(str),
                keyring_filename=IsInstance(str),
                keyring_data=IsInstance(str),  # ??? Binary, no?
                created=IsInstance(datetime),
                updated=IsInstance(datetime),
            )),
        )


# TestBootSourceSelections
# TestDevices


class TestEvents(IntegrationTestCase):

    def test__query_events(self):
        events = self.origin.Events.query()
        self.assertThat(events, IsInstance(self.origin.Events))
        events = events.prev()
        self.assertThat(events, IsInstance(self.origin.Events))
        events = events.next()
        self.assertThat(events, IsInstance(self.origin.Events))

    def test__events(self):
        self.assertThat(
            self.origin.Events.query(),
            AllMatch(MatchesStructure(
                event_id=IsInstance(int),
                event_type=IsInstance(str),
                system_id=IsInstance(str),
                hostname=IsInstance(str),
                level=IsInstance(viscera.events.Level),
                created=IsInstance(datetime),
                description=IsInstance(str),
            )),
        )


# TestFiles
# TestMAAS


class TestMachines(IntegrationTestCase):

    def test__list_machines(self):
        machines = self.origin.Machines.read()
        self.assertThat(machines, MatchesAll(
            IsInstance(self.origin.Machines),
            AllMatch(IsInstance(self.origin.Machine)),
        ))
        self.assertThat(
            machines,
            AllMatch(MatchesStructure(
                # This is NOT exhaustive.
                system_id=IsInstance(str),
                architecture=IsInstance(str),
                hostname=IsInstance(str),
                ip_addresses=IsInstance(list),
                status=IsInstance(int),
                status_name=IsInstance(str),
                tags=IsInstance(list),
            )),
        )

    def XXXtest__allocate_deploy_and_release(self):
        machines_ready = [
            machine for machine in self.origin.Machines.read()
            if machine.status_name == "Ready"
        ]
        if len(machines_ready) == 0:
            self.skip("No machines available.")

        # Allocate one of the ready machines. XXX: This ought to be a method
        # on Machine or take a `system_id` argument.
        machine = random.choice(machines_ready)
        machine = self.origin.Machines.allocate(hostname=machine.hostname)
        self.assertThat(machine.status_name, Equals("Allocated"))

        try:
            # Deploy the machine with defaults.
            machine = machine.deploy()
            self.assertThat(machine.status_name, Equals("Deploying"))
            # Wait for the machine to deploy.
            for elapsed, remaining, wait in retries(600, 10):
                machine = self.origin.Machine.read(machine.system_id)
                if machine.status_name == "Deploying":
                    sleep(wait)
                else:
                    break
            else:
                self.fail("Timed-out waiting for machine to deploy.")
            # The machine has deployed.
            self.assertThat(machine.status_name, Equals("Deployed"))

        finally:
            # Release the machine.
            machine = machine.release("Finished with this now, thanks.")
            self.assertThat(machine.status_name, Equals("Releasing"))
            # Wait for the machine to release.
            for elapsed, remaining, wait in retries(300, 10):
                machine = self.origin.Machine.read(machine.system_id)
                if machine.status_name == "Releasing":
                    sleep(wait)
                else:
                    break
            else:
                self.fail("Timed-out waiting for machine to release.")
            # The machine has been released.
            self.assertThat(machine.status_name, Equals("Ready"))


class TestRackControllers(IntegrationTestCase):

    def test__list_rack_controllers(self):
        machines = self.origin.RackControllers.read()
        self.assertThat(machines, MatchesAll(
            IsInstance(self.origin.RackControllers),
            AllMatch(IsInstance(self.origin.RackController)),
        ))
        self.assertThat(
            machines,
            AllMatch(MatchesStructure(
                # This is NOT exhaustive.
                architecture=IsInstance(str),
                cpus=IsInstance(int),
                distro_series=IsInstance(str),
                fqdn=IsInstance(str),
                hostname=IsInstance(str),
                ip_addresses=IsInstance(list),
                memory=IsInstance(int),
                power_state=IsInstance(str),
                system_id=IsInstance(str),
                zone=IsInstance(self.origin.Zone),
            )),
        )


class TestRegionControllers(IntegrationTestCase):

    def test__list_region_controllers(self):
        machines = self.origin.RegionControllers.read()
        self.assertThat(machines, MatchesAll(
            IsInstance(self.origin.RegionControllers),
            AllMatch(IsInstance(self.origin.RegionController)),
        ))
        self.assertThat(
            machines,
            AllMatch(MatchesStructure(
                # This is NOT exhaustive.
                architecture=IsInstance(str),
                cpus=IsInstance(int),
                distro_series=IsInstance(str),
                fqdn=IsInstance(str),
                hostname=IsInstance(str),
                ip_addresses=IsInstance(list),
                memory=IsInstance(int),
                power_state=IsInstance(str),
                system_id=IsInstance(str),
                zone=IsInstance(self.origin.Zone),
            )),
        )


# TestTags
# TestTesting
# TestUsers
# TestVersion
# TestZones


# Additional matchers.

def Optional(matcher, default=Is(None)):
    return MatchesAny(matcher, default)


# End.



"""Test for `maas.client.viscera.block_devices`."""

import random

from testtools.matchers import Equals

from ..block_devices import BlockDevice, BlockDevices
from ..nodes import Node, Nodes
from ..partitions import Partition, Partitions

from ..testing import bind
from ...enum import BlockDeviceType
from ...testing import make_string_without_spaces, TestCase


def make_origin():
    """
    Create a new origin with required objects.
    """
    return bind(BlockDevices, BlockDevice, Partitions, Partition, Nodes, Node)


class TestBlockDevices(TestCase):
    def test__by_name(self):
        origin = make_origin()
        BlockDevices, BlockDevice = origin.BlockDevices, origin.BlockDevice
        system_id = make_string_without_spaces()
        block_names = [make_string_without_spaces() for _ in range(3)]
        blocks_by_name = {
            name: BlockDevice(
                {
                    "system_id": system_id,
                    "id": random.randint(0, 100),
                    "name": name,
                    "type": BlockDeviceType.PHYSICAL.value,
                }
            )
            for name in block_names
        }
        blocks = BlockDevices([obj for _, obj in blocks_by_name.items()])
        self.assertEqual(blocks_by_name, blocks.by_name)

    def test__get_by_name(self):
        origin = make_origin()
        BlockDevices, BlockDevice = origin.BlockDevices, origin.BlockDevice
        system_id = make_string_without_spaces()
        block_names = [make_string_without_spaces() for _ in range(3)]
        blocks_by_name = {
            name: BlockDevice(
                {
                    "system_id": system_id,
                    "id": random.randint(0, 100),
                    "name": name,
                    "type": BlockDeviceType.PHYSICAL.value,
                }
            )
            for name in block_names
        }
        bcaches = BlockDevices([obj for _, obj in blocks_by_name.items()])
        name = block_names[0]
        self.assertEqual(blocks_by_name[name], bcaches.get_by_name(name))

    def test__read_bad_node_type(self):
        BlockDevices = make_origin().BlockDevices
        error = self.assertRaises(TypeError, BlockDevices.read, random.randint(0, 100))
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__read_with_system_id(self):
        BlockDevices = make_origin().BlockDevices
        system_id = make_string_without_spaces()
        blocks = [
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "name": make_string_without_spaces(),
                "type": BlockDeviceType.PHYSICAL.value,
            }
            for _ in range(3)
        ]
        BlockDevices._handler.read.return_value = blocks
        blocks = BlockDevices.read(system_id)
        self.assertThat(len(blocks), Equals(3))
        BlockDevices._handler.read.assert_called_once_with(system_id=system_id)

    def test__read_with_Node(self):
        origin = make_origin()
        BlockDevices, Node = origin.BlockDevices, origin.Node
        system_id = make_string_without_spaces()
        node = Node(system_id)
        blocks = [
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "name": make_string_without_spaces(),
                "type": BlockDeviceType.PHYSICAL.value,
            }
            for _ in range(3)
        ]
        BlockDevices._handler.read.return_value = blocks
        blocks = BlockDevices.read(node)
        self.assertThat(len(blocks), Equals(3))
        BlockDevices._handler.read.assert_called_once_with(system_id=system_id)

    def test__create_bad_node_type(self):
        origin = make_origin()
        BlockDevices = origin.BlockDevices
        error = self.assertRaises(
            TypeError,
            BlockDevices.create,
            random.randint(0, 100),
            "sda",
            model="QEMU",
            serial="QEMU0001",
            size=(4096 * 1024),
            block_size=512,
        )
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__create_missing_size(self):
        BlockDevices = make_origin().BlockDevices
        error = self.assertRaises(
            ValueError,
            BlockDevices.create,
            make_string_without_spaces(),
            "sda",
            model="QEMU",
            serial="QEMU0001",
            block_size=512,
        )
        self.assertEquals("size must be provided and greater than zero.", str(error))

    def test__create_negative_size(self):
        BlockDevices = make_origin().BlockDevices
        error = self.assertRaises(
            ValueError,
            BlockDevices.create,
            make_string_without_spaces(),
            "sda",
            model="QEMU",
            serial="QEMU0001",
            size=-1,
            block_size=512,
        )
        self.assertEquals("size must be provided and greater than zero.", str(error))

    def test__create_missing_block_size(self):
        BlockDevices = make_origin().BlockDevices
        error = self.assertRaises(
            ValueError,
            BlockDevices.create,
            make_string_without_spaces(),
            "sda",
            model="QEMU",
            serial="QEMU0001",
            size=(4096 * 1024),
            block_size=None,
        )
        self.assertEquals(
            "block_size must be provided and greater than zero.", str(error)
        )

    def test__create_negative_block_size(self):
        BlockDevices = make_origin().BlockDevices
        error = self.assertRaises(
            ValueError,
            BlockDevices.create,
            make_string_without_spaces(),
            "sda",
            model="QEMU",
            serial="QEMU0001",
            size=(4096 * 1024),
            block_size=-1,
        )
        self.assertEquals(
            "block_size must be provided and greater than zero.", str(error)
        )

    def test__create_model_requires_serial(self):
        BlockDevices = make_origin().BlockDevices
        error = self.assertRaises(
            ValueError,
            BlockDevices.create,
            make_string_without_spaces(),
            "sda",
            model="QEMU",
            size=(4096 * 1024),
        )
        self.assertEquals("serial must be provided when model is provided.", str(error))

    def test__create_serial_requires_model(self):
        BlockDevices = make_origin().BlockDevices
        error = self.assertRaises(
            ValueError,
            BlockDevices.create,
            make_string_without_spaces(),
            "sda",
            serial="QEMU0001",
            size=(4096 * 1024),
        )
        self.assertEquals("model must be provided when serial is provided.", str(error))

    def test__create_requires_model_and_serial_or_id_path(self):
        BlockDevices = make_origin().BlockDevices
        error = self.assertRaises(
            ValueError,
            BlockDevices.create,
            make_string_without_spaces(),
            "sda",
            size=(4096 * 1024),
        )
        self.assertEquals(
            "Either model/serial is provided or id_path must be provided.", str(error)
        )

    def test__create(self):
        origin = make_origin()
        BlockDevices = origin.BlockDevices
        system_id = make_string_without_spaces()
        size = 4096 * 1024
        BlockDevices._handler.create.return_value = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "name": "sda",
            "model": "QEMU",
            "serial": "QEMU0001",
            "size": size,
            "block_size": 512,
        }
        BlockDevices.create(
            system_id, "sda", model="QEMU", serial="QEMU0001", size=size
        )
        BlockDevices._handler.create.assert_called_once_with(
            system_id=system_id,
            name="sda",
            model="QEMU",
            serial="QEMU0001",
            size=size,
            block_size=512,
        )

    def test__create_with_tags(self):
        origin = make_origin()
        BlockDevices, BlockDevice = origin.BlockDevices, origin.BlockDevice
        system_id = make_string_without_spaces()
        block_id = random.randint(0, 100)
        size = 4096 * 1024
        BlockDevices._handler.create.return_value = {
            "system_id": system_id,
            "id": block_id,
            "name": "sda",
            "model": "QEMU",
            "serial": "QEMU0001",
            "size": size,
            "block_size": 512,
            "tags": [],
        }
        BlockDevices._handler.add_tag.return_value = None
        tag = make_string_without_spaces()
        BlockDevices.create(
            system_id, "sda", model="QEMU", serial="QEMU0001", size=size, tags=[tag]
        )
        BlockDevices._handler.create.assert_called_once_with(
            system_id=system_id,
            name="sda",
            model="QEMU",
            serial="QEMU0001",
            size=size,
            block_size=512,
        )
        BlockDevice._handler.add_tag.assert_called_once_with(
            system_id=system_id, id=block_id, tag=tag
        )


class TestBlockDevice(TestCase):
    def test__read_bad_node_type(self):
        BlockDevice = make_origin().BlockDevice
        error = self.assertRaises(
            TypeError, BlockDevice.read, random.randint(0, 100), random.randint(0, 100)
        )
        self.assertEquals("node must be a Node or str, not int", str(error))

    def test__read_with_system_id(self):
        BlockDevice = make_origin().BlockDevice
        system_id = make_string_without_spaces()
        block = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "type": BlockDeviceType.PHYSICAL.value,
        }
        BlockDevice._handler.read.return_value = block
        BlockDevice.read(system_id, block["id"])
        BlockDevice._handler.read.assert_called_once_with(
            system_id=system_id, id=block["id"]
        )

    def test__read_with_Node(self):
        origin = make_origin()
        BlockDevice = origin.BlockDevice
        Node = origin.Node
        system_id = make_string_without_spaces()
        node = Node(system_id)
        block = {
            "system_id": system_id,
            "id": random.randint(0, 100),
            "type": BlockDeviceType.PHYSICAL.value,
        }
        BlockDevice._handler.read.return_value = block
        BlockDevice.read(node, block["id"])
        BlockDevice._handler.read.assert_called_once_with(
            system_id=system_id, id=block["id"]
        )

    def test__save_add_tag(self):
        BlockDevice = make_origin().BlockDevice
        system_id = make_string_without_spaces()
        block = BlockDevice(
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "type": BlockDeviceType.PHYSICAL.value,
                "tags": [],
            }
        )
        tag = make_string_without_spaces()
        block.tags.append(tag)
        BlockDevice._handler.add_tag.return_value = None
        block.save()
        BlockDevice._handler.add_tag.assert_called_once_with(
            system_id=system_id, id=block.id, tag=tag
        )

    def test__save_remove_tag(self):
        BlockDevice = make_origin().BlockDevice
        system_id = make_string_without_spaces()
        tag = make_string_without_spaces()
        block = BlockDevice(
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "type": BlockDeviceType.PHYSICAL.value,
                "tags": [tag],
            }
        )
        block.tags.remove(tag)
        BlockDevice._handler.remove_tag.return_value = None
        block.save()
        BlockDevice._handler.remove_tag.assert_called_once_with(
            system_id=system_id, id=block.id, tag=tag
        )

    def test__delete(self):
        BlockDevice = make_origin().BlockDevice
        system_id = make_string_without_spaces()
        block = BlockDevice(
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "type": BlockDeviceType.PHYSICAL.value,
            }
        )
        BlockDevice._handler.delete.return_value = None
        block.delete()
        BlockDevice._handler.delete.assert_called_once_with(
            system_id=system_id, id=block.id
        )

    def test__set_as_boot_disk(self):
        BlockDevice = make_origin().BlockDevice
        system_id = make_string_without_spaces()
        block = BlockDevice(
            {
                "system_id": system_id,
                "id": random.randint(0, 100),
                "type": BlockDeviceType.PHYSICAL.value,
            }
        )
        BlockDevice._handler.set_boot_disk.return_value = None
        block.set_as_boot_disk()
        BlockDevice._handler.set_boot_disk.assert_called_once_with(
            system_id=system_id, id=block.id
        )

    def test__format(self):
        BlockDevice = make_origin().BlockDevice
        system_id = make_string_without_spaces()
        block_id = random.randint(0, 100)
        block = BlockDevice(
            {
                "system_id": system_id,
                "id": block_id,
                "type": BlockDeviceType.PHYSICAL.value,
            }
        )
        BlockDevice._handler.format.return_value = {
            "system_id": system_id,
            "id": block_id,
            "type": BlockDeviceType.PHYSICAL.value,
            "filesystem": {"fstype": "ext4"},
        }
        block.format("ext4")
        BlockDevice._handler.format.assert_called_once_with(
            system_id=system_id, id=block_id, fstype="ext4", uuid=None
        )

    def test__unformat(self):
        BlockDevice = make_origin().BlockDevice
        system_id = make_string_without_spaces()
        block_id = random.randint(0, 100)
        block = BlockDevice(
            {
                "system_id": system_id,
                "id": block_id,
                "type": BlockDeviceType.PHYSICAL.value,
                "filesystem": {"fstype": "ext4"},
            }
        )
        BlockDevice._handler.unformat.return_value = {
            "system_id": system_id,
            "id": block_id,
            "type": BlockDeviceType.PHYSICAL.value,
            "filesystem": None,
        }
        block.unformat()
        BlockDevice._handler.unformat.assert_called_once_with(
            system_id=system_id, id=block_id
        )

    def test__mount(self):
        BlockDevice = make_origin().BlockDevice
        system_id = make_string_without_spaces()
        block_id = random.randint(0, 100)
        block = BlockDevice(
            {
                "system_id": system_id,
                "id": block_id,
                "type": BlockDeviceType.PHYSICAL.value,
                "filesystem": {"fstype": "ext4"},
            }
        )
        BlockDevice._handler.mount.return_value = {
            "system_id": system_id,
            "id": block_id,
            "type": BlockDeviceType.PHYSICAL.value,
            "filesystem": {
                "fstype": "ext4",
                "mount_point": "/",
                "mount_options": "noatime",
            },
        }
        block.mount("/", mount_options="noatime")
        BlockDevice._handler.mount.assert_called_once_with(
            system_id=system_id, id=block_id, mount_point="/", mount_options="noatime"
        )

    def test__unmount(self):
        BlockDevice = make_origin().BlockDevice
        system_id = make_string_without_spaces()
        block_id = random.randint(0, 100)
        block = BlockDevice(
            {
                "system_id": system_id,
                "id": block_id,
                "type": BlockDeviceType.PHYSICAL.value,
                "filesystem": {
                    "fstype": "ext4",
                    "mount_point": "/",
                    "mount_options": "noatime",
                },
            }
        )
        BlockDevice._handler.unmount.return_value = {
            "system_id": system_id,
            "id": block_id,
            "type": BlockDeviceType.PHYSICAL.value,
            "filesystem": {"fstype": "ext4", "mount_point": "", "mount_options": ""},
        }
        block.unmount()
        BlockDevice._handler.unmount.assert_called_once_with(
            system_id=system_id, id=block_id
        )



"""Objects for events."""

__all__ = ["Events"]

from datetime import datetime
import enum
from functools import partial
import logging
import typing
from urllib.parse import parse_qs, urlparse
from .users import User

import pytz

from . import Object, ObjectField, ObjectSet, ObjectType
from ..utils.maas_async import is_loop_running

#
# The query API call returns:
#
#     dict(
#         count=displayed_events_count,
#         events=[...],
#         next_uri=next_uri,
#         prev_uri=prev_uri,
#     )
#
# An event looks like:
#
#     dict(
#         node=event.node.system_id,
#         hostname=event.node.hostname,
#         id=event.id,
#         level=event.type.level_str,
#         created=event.created.strftime('%a, %d %b. %Y %H:%M:%S'),
#         type=event.type.description,
#         description=event.description,
#         username=event.user.username
#     )
#
# Notes:
# 1. displayed_events_count is len(events). In other words, superfluous.
#


class Level(enum.IntEnum):
    """Logging levels used in MAAS.

    They happen to correspond to levels in the `logging` module.
    """

    AUDIT = 0
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL

    @classmethod
    def normalise(cls, level):
        """Convert the given level into a `Level`.

        :param level: A `Level` instance, a string that matches the name of a
            level, or an integer matching the value of a level.
        :raise ValueError: If the level cannot be found.
        """
        try:
            return cls[level]
        except KeyError:
            return cls(level)


class EventsType(ObjectType):
    """Metaclass for `Events`."""

    Level = Level

    async def query(
        cls,
        *,
        hostnames: typing.Iterable[str] = None,
        domains: typing.Iterable[str] = None,
        zones: typing.Iterable[str] = None,
        macs: typing.Iterable[str] = None,
        system_ids: typing.Iterable[str] = None,
        agent_name: str = None,
        level: typing.Union[Level, int, str] = None,
        before: int = None,
        after: int = None,
        limit: int = None,
        owner: typing.Union[User, str] = None
    ):
        """Query MAAS for matching events."""

        if before is not None and after is not None:
            raise ValueError("Specify either `before` or `after`, not both.")

        params = {}

        if hostnames is not None:
            params["hostname"] = list(hostnames)
        if domains is not None:
            params["domain"] = list(domains)
        if zones is not None:
            params["zone"] = list(zones)
        if macs is not None:
            params["mac_address"] = list(macs)
        if system_ids is not None:
            params["id"] = list(system_ids)
        if agent_name is not None:
            params["agent_name"] = [agent_name]
        if level is not None:
            level = Level.normalise(level)
            params["level"] = [level.name]
        if before is not None:
            params["before"] = ["{:d}".format(before)]
        if after is not None:
            params["after"] = ["{:d}".format(after)]
        if limit is not None:
            params["limit"] = ["{:d}".format(limit)]
        if owner is not None:
            if isinstance(owner, User):
                params["owner"] = [owner.username]
            elif isinstance(owner, str):
                params["owner"] = [owner]
            else:
                raise TypeError(
                    "owner must be either User or str, not %s" % (type(owner).__name__)
                )

        data = await cls._handler.query(**params)
        return cls(data)


def parse_created_timestamp(created):
    created = datetime.strptime(created, "%a, %d %b. %Y %H:%M:%S")
    return created.replace(tzinfo=pytz.UTC)


class Events(ObjectSet, metaclass=EventsType):
    """MAAS event information."""

    __slots__ = "_prev_uri", "_next_uri"

    def __init__(self, eventmap):
        events = map(self._object, eventmap["events"])
        super(Events, self).__init__(events)
        self._prev_uri = eventmap["prev_uri"]
        self._next_uri = eventmap["next_uri"]

    @classmethod
    async def _fetch(cls, uri, count):
        query = urlparse(uri).query
        params = parse_qs(query, errors="strict")
        if count is not None:
            params["limit"] = ["{:d}".format(count)]
        data = await cls._handler.query(**params)
        return cls(data)

    async def prev(self, count=None):
        """Load the previous (older) page/batch of events.

        :param count: A limit on the number of events to fetch. By default the
            limit is the same as that specified for this set of events.
        """
        if self._prev_uri is None:
            return self[:0]  # An empty slice of `self`.
        else:
            return self._fetch(self._prev_uri, count)

    async def next(self, count=None):
        """Load the next (newer) page/batch of events.

        :param count: A limit on the number of events to fetch. By default the
            limit is the same as that specified for this set of events.
        """
        if self._next_uri is None:
            return self[:0]  # An empty slice of `self`.
        else:
            return self._fetch(self._next_uri, count)

    def backwards(self):
        """Iterate continuously through events, backwards.

        Note: this returns an *asynchronous* iterator.

        This will load new pages/batches of events on demand.
        """
        if is_loop_running():
            return EventsAsyncIteratorBackwards(self)
        else:
            return self._backwards_sync()

    def _backwards_sync(self):
        current = self
        while len(current) != 0:
            yield from current
            if is_loop_running():
                raise RuntimeError(
                    "Cannot iterate synchronously while " "event-loop is running."
                )
            current = current.prev()

    def forwards(self):
        """Iterate continuously through events, forwards.

        Note: this returns an *asynchronous* iterator.

        This will load new pages/batches of events on demand.
        """
        if is_loop_running():
            return EventsAsyncIteratorForwards(self)
        else:
            return self._forwards_sync()

    def _forwards_sync(self):
        current = self
        while len(current) != 0:
            yield from reversed(current)
            if is_loop_running():
                raise RuntimeError(
                    "Cannot iterate synchronously while " "event-loop is running."
                )
            current = current.next()


class EventsAsyncIteratorBackwards:
    def __init__(self, current):
        super(EventsAsyncIteratorBackwards, self).__init__()
        self._current_iter = iter(current)
        self._current = current

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self._current_iter)
        except StopIteration:
            self._current = await self._current.prev()
            self._current_iter = iter(self._current)
            if len(self._current) == 0:
                raise StopAsyncIteration()
            else:
                return next(self._current_iter)


class EventsAsyncIteratorForwards:
    def __init__(self, current):
        super(EventsAsyncIteratorForwards, self).__init__()
        self._current_iter = reversed(current)
        self._current = current

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self._current_iter)
        except StopIteration:
            self._current = await self._current.next()
            self._current_iter = reversed(self._current)
            if len(self._current) == 0:
                raise StopAsyncIteration()
            else:
                return next(self._current_iter)


def truncate(length, text):  # TODO: Move into utils.
    """Truncate the given text to the given length.

    If the text is longer than the ``length``, return a new string that
    contains ``length - 1`` characters of text followed by an ellipsis.
    Otherwise return the given text unaltered.
    """
    if len(text) > length:
        return text[: length - 1] + "…"
    else:
        return text


class Event(Object):
    """An event."""

    event_id = ObjectField("id", readonly=True)
    event_type = ObjectField("type", readonly=True)

    system_id = ObjectField("node", readonly=True)
    hostname = ObjectField("hostname", readonly=True)

    level = ObjectField.Checked("level", Level.normalise, readonly=True)
    created = ObjectField.Checked("created", parse_created_timestamp, readonly=True)

    description = ObjectField("description", readonly=True)
    description_short = ObjectField.Checked(
        "description", partial(truncate, 50), readonly=True
    )

    username = ObjectField("username", readonly=True)

    def __repr__(self):
        return (
            "<{self.__class__.__name__} {self.created:%Y-%m-%d %H:%M:%S} "
            "{self.level.name} {self.hostname} {self.description_short!r}>"
        ).format(self=self)
